/*

    -------------------------------------------------------

    ADAPTED FROM THE BELOW
    Jul 11, 2001
    Theo Verelst
    http://theover.tripod.com , theover@yahoo.com

    This is the main file to link the string simulator Pms
    with the portaudio routines to do live simulations.

    -------------------------------------------------------

 * patest_saw.c
 * Play a simple sawtooth wave.
 *
 * Author: Phil Burk  http://www.softsynth.com
 *
 * This program uses the PortAudio Portable Audio Library.
 * For more information see: http://www.portaudio.com
 * Copyright (c) 1999-2000 Ross Bencina and Phil Burk
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * Any person wishing to distribute modifications to the Software is
 * requested to send the modifications to the original developer so that
 * they can be incorporated into the canonical version.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <stdio.h>
#include <math.h>
#include "portaudio.h"
#define NUM_SECONDS   (8)
#define SAMPLE_RATE   (44100)

PortAudioStream *stream;

typedef struct {
	float left_phase;
	float right_phase;
}paTestDataf;

typedef struct {
	short left_phase;
	short right_phase;
}paTestData;

extern int n[];

/* This routine will be called by the PortAudio engine when audio is needed.
** It may called at interrupt level on some machines so don't do anything
** that could mess up the system like calling malloc() or free().
*/
static int patestCallback(	void *inputBuffer, void *outputBuffer,
				unsigned long framesPerBuffer,
				PaTimestamp outTime, void *userData )
{
/* Cast data passed through stream to our structure. */
	paTestData *data = (paTestData*)userData;
	short *out = (short*)outputBuffer;
	unsigned int i;
	(void) outTime; /* Prevent unused variable warnings. */
	(void) inputBuffer;

        simulate(n,framesPerBuffer,out);

/*	
	for( i=0; i<framesPerBuffer; i++ )
	{
		*out++ = data->left_phase;		
		*out++ = data->right_phase;
		data->left_phase += 400;
		if( data->left_phase >= 20000 ) data->left_phase -= 40000;
		data->right_phase += 401; 
		if( data->right_phase >= 20000 ) data->right_phase -= 40000;
	}
*/
	return 0;
}
/*******************************************************************/
static paTestData data;
int pa_maini(int dt)
{
	PaError err;
	printf("PMS / PortAudio Test: output simulated guitar string wave.\n");
/* Initialize our data for use by callback. */
	data.left_phase = data.right_phase = 0;
/* Initialize library before making any other calls. */
	err = Pa_Initialize();
	if( err != paNoError ) goto errori;
/* Open an audio I/O stream. */
	err = Pa_OpenDefaultStream(
				&stream,
				0,              /* no input channels */
				2,              /* stereo output */
				paInt16,      /* 16 bit int output */
				SAMPLE_RATE,
				128,            /* frames per buffer */
				2,              /* number of buffers, if zero then use default minimum */
				patestCallback,
				&data );
	if( err != paNoError ) goto errori;
	err = Pa_StartStream( stream );
	if( err != paNoError ) goto errori;
/* Sleep for several seconds. */
/*	Pa_Sleep(dt*1000); */
	return(err);
errori:
	Pa_Terminate();
	fprintf( stderr, "An error occured while using the portaudio stream\n" ); 
	fprintf( stderr, "Error number: %d\n", err );
	fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
	return err;
}

int pa_maine()
{
	PaError err;
	err = Pa_StopStream( stream );
	if( err != paNoError ) goto errore;
	err = Pa_CloseStream( stream );
	if( err != paNoError ) goto errore;
	Pa_Terminate();
	printf("Test finished.\n");
	return err;
errore:
	Pa_Terminate();
	fprintf( stderr, "An error occured while using the portaudio stream\n" ); 
	fprintf( stderr, "Error number: %d\n", err );
	fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
	return err;
}
