/************* wave.c: Basic Waveguide Code *****************/
/* 
    sound lib is seperated, wav support is in, as is midi support 
    contains main() and argument processing.
*/


/************************************************************************************
   COPYRIGHT

  All rights are reserved by the author, who can be reached through 
      theover@yahoo.com (or see http://theover.tripod.com)
  exept that permission is granted for non-commercial use 
  of sources or programs and the sounds coming from them, 
  though the rights on the algorithms are strictly reserved.
  The author relies on the courtesy of users or programmers 
  who make small or large adaptations to inform him about them, 
  and demands that every copy or adaptation, also partial, 
  clearly contains this copyright notice.

*************************************************************************************/

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include <windows.h>
#include <mmsystem.h>

#include <sys/types.h>

/* define PREMUL 32768           * Fixed exponent  */
#define PREMUL 5000           /* Fixed exponent  */
#define MAXN   8192            /* Max # nodes     */
#define SRATE  44100           /* Wav Sample Rate */
#define MAXNSTRINGS 12
#define NSTRINGS 3
#ifndef PI
#define PI 3.1415926535
#endif

int notcl = 0, nosound = 0;    /* to prevent (time expensive) output */

int e[MAXNSTRINGS][MAXN], v[MAXNSTRINGS][MAXN];          /* displacement and velocity arrays   */
int u1 =1024, u2 =1024;        /* combined values of (c/m)*PREMUL*dt */
int n[MAXNSTRINGS];                     /* number of nodes                    */
int midinote[MAXNSTRINGS];
int vel[MAXNSTRINGS];
double bend = 0.0;
double amp = 1.0;
int steps;                     /* number of (full wave) iter. steps  */
int fd;                        /* file descriptor for sound file     */
int binoutnode;                /* node to output in sound file       */
FILE *tclf;                    /* socket driven tcl file (ptr)       */
int s, uc;
int nstrings;
int volume;

HMIDIIN inHandle;

extern pick_main(int *argc, char **argv);
extern void pick_loop();
extern void pick_end();
extern void volumechanged();

/*
extern int do_read();
extern char in[], line[];
extern int serv_main();
*/

char line[4*1024];
int wait_read() {
}
int do_read() {
gets(line);
}
int serv_main() { }
int do_write() { }

silenceone(w)
int w;
{
   int i;
   for (i=0; i<n[w]; i++) {
      e[w][i] = 0; v[w][i] = 0;      /* init to zero energy / strain       */
   }
   
}

/* w=-1 sound all; w=-2 silence all; w = string: sound string */

init(n,w)
int n[], w;
{
   int i;
  for (s=0; s<nstrings; s++) 
  if (w==-2 || w==s) {
   for (i=0; i<n[s]; i++) {
      e[s][i] = 0; v[s][i] = 0;      /* init to zero energy / strain       */
   }
  }
  for (s=0; s<nstrings; s++) 
  if ((w<0 && w!=-2) || w==s) {
   for (i=1; i<n[s]-1; i++) {
      /* v[s][i] = (int) (PREMUL* 1000 *  */
      v[s][i] = (int) (PREMUL* 2* vel[s] * 
                (cos(
                   (1.5*1)* 2.0 * PI * (double) i/ (double) (n[s]-2)) 
                 ) 
                 /exp(
                   pow(
                      /* ((512.0/128.0) * (double) (i-n[s]/1.5) */
                      (((double) vel[s]/128.0) * (double) (i-n[s]/1.3)
                         /(double) n[s]
                      ),2)
                 ));
      e[s][i] = -v[s][i]/2; /* (int) (PREMUL/2 * cos(2.0 * PI * (double) i/ (double) n) ); */
   }
  }
   return;

}



inittcl(n)
int n[];
{
   int i;
   if (notcl) return;
   printf("if {[winfo exists .c] == 0} {canvas .c; pack .c -expand y -fill both}\n");
   printf("proc graphb {n} {global a; for {set i 0} {$i<$n} {set i [expr $i+1]} {.c create line [expr 2*$i] 300 [expr 2*$i] [expr 300-$a($i)/30] -width 1 -fill red}\n.c create line 0 300 [expr 2* $n] 300 -fill green}\n");
   printf("proc graph {n} {global a; for {set i 1} {$i<$n} {set i [expr $i+1]} {.c create line [expr 2*$i] [expr 300-$a([expr $i-1])/30] [expr 2*$i] [expr 300-$a($i)/30] -width 1 -fill purple}\n.c create line 0 300 [expr 2* $n] 300 -fill blue}\n");
   printf("set nnodes %d\n",n[0]);
   printf(".c delete all\n");
   for (i=0; i<n[0]; i++)
      printf("set a(%d) %d\n",i,e[0][i]/PREMUL);
   printf("graphb $nnodes\n update \n after 2000\nset vt 2\n");
}

initbin(n,s)                     /* could add a .wav header here   */
int n;
char *s;
{
   if (nosound) return;
   fd = open(s,O_RDWR|O_CREAT|O_TRUNC|O_BINARY);
}

iwrite(fd,n,l)
int fd,l;
unsigned int n;
{
   write(fd,&n,l);
}

initwav(n,s)                     /* Proto wav header, n=nnodes, filename */
int n;
char *s;
{
   if (nosound) return;
   fd = open(s,O_RDWR|O_CREAT|O_TRUNC|O_BINARY);
   if ((void *) fd == NULL ) return;

   write(fd,"RIFF",4);
   iwrite(fd,(2*steps+36),4);
   write(fd,"WAVE",4);

   write(fd,"fmt ",4);
   iwrite(fd,(0x10),4);
   iwrite(fd,((short) 0x01),2);
   iwrite(fd,((short) 1),2);              /* Mono */
   iwrite(fd,(SRATE),4);                  /* Sample rate */
   iwrite(fd,(2*SRATE),4);
   iwrite(fd,((short) 2),2);
   iwrite(fd,((short) 16),2);

   write(fd,"data",4);
   iwrite(fd,(2*steps),4);
}


outputtcl(n)
int n[];
{
   int i;
   if (notcl) return;
   for (i=0; i<n[0]; i++)
      printf("set a%d(%d) %d\n",i,e[0][i]/PREMUL);
   printf(".c delete all\n");
   printf("graph $nnodes\n update\n after $vt\n");
}

foutputtcl(n)
int n[];
{
   int i;
   if (tclf == NULL) return;
   fprintf(tclf,"set nnodes %d\n",n[0]);
   for (i=0; i<n[0]; i++)
      fprintf(tclf,"set a(%d) %d\n",i,e[0][i]/PREMUL);
   fprintf(tclf,"graphdel\n");
   fprintf(tclf,"graph $nnodes\nupdate\nafter $vt\n");
}

outputbin(i)
int i;
{
   char *cp;
   int t;
   if (fd > 0) {
    t = 0;
    for (s=0; s<nstrings; s++) {
      t += e[s][i]/(PREMUL);
    }
      cp = (char *) &t;
      write(fd,cp,1);
      write(fd,cp+1,1);
   }
}

outputpa(i,b)
int i;
short *b;
{
   char *cp;
   int l,r;
    l = r = 0;
/*    for (s=0; s<nstrings+0; s++) {
      r += (s+1) * e[s][i]/(PREMUL);
      l += (1+nstrings-s) * e[s][i]/(PREMUL);
    } */
    for (s=0; s<nstrings+0; s++) {
      r += (nstrings/2+1) * e[s][i]/(PREMUL);
      l += (1+nstrings-nstrings/2) * e[s][i]/(PREMUL);
    }
    *b++ = (short) (l/(nstrings+2)) ;
    *b   = (short) (r/(nstrings+2)) ;
}


propagate_fixend(n)              /* Compute new velocity and           */
int n[];                         /* displacements once for all nodes.  */
{                                /* (more optimization possible here)  */
   register int i, tt;
  for (s=0; s<nstrings; s++) {
  if (uc>s*4*800) {
   for (i=1; i<n[s]-1; i++) {
      /* v[i] += (u1 * (e[i-1]+e[i+1]-2*e[i]) )/PREMUL; */
      v[s][i] += e[s][i-1];
      v[s][i] += e[s][i+1];
      v[s][i] -= e[s][i];
      v[s][i] -= e[s][i];
   }
   for (i=1; i<n[s]-1; i++) {
      /* e[i] += (u2 * v[i] ) / PREMUL; */
      e[s][i] += v[s][i];
   }
/*      e[1] *= 5;
      e[1] /= 10;
      e[n-2] *= 5;
      e[n-2] /= 10; */

      if (midinote[s] >0) {
      e[s][1] *= (95-0);         /* more damping for higher string numbers */
      e[s][1] /= 100;
      e[s][1] = (int) (  (double) 15000.0* (double) PREMUL * atan(
            ((double) e[s][1]) / ((double) PREMUL * (double) 15000.0)
         ));
      e[s][n[s]-2] *= (95-0);
      e[s][n[s]-2] /= 100;
      e[s][n[s]-2] = (int) ( (double) 10000.0* (double) PREMUL * atan(
            ((double) e[s][n[s]-2]) / ((double) PREMUL * (double) 10000.0)
         ));
      e[s][2] *= (96);
      e[s][2] /= 100;
      e[s][n[s]-3] *= 97;
      e[s][n[s]-3] /= 100;
      e[s][3] *= (98);
      e[s][3] /= 100;
      e[s][n[s]-4] *= 98;
      e[s][n[s]-4] /= 100;
      e[s][4] *= (99);
      e[s][4] /= 100;
      e[s][n[s]-5] *= 99;
      e[s][n[s]-5] /= 100;
      } else {
      e[s][1] *= (89-0);         /* more damping for higher string numbers */
      e[s][1] /= 100;
      e[s][1] = (int) (  (double) 15000.0* (double) PREMUL * atan(
            ((double) e[s][1]) / ((double) PREMUL * (double) 15000.0)
         ));
      e[s][n[s]-2] *= (89-0);
      e[s][n[s]-2] /= 100;
      e[s][n[s]-2] = (int) ( (double) 10000.0* (double) PREMUL * atan(
            ((double) e[s][n[s]-2]) / ((double) PREMUL * (double) 10000.0)
         ));
      e[s][2] *= (90);
      e[s][2] /= 100;
      e[s][n[s]-3] *= 90;
      e[s][n[s]-3] /= 100;
      e[s][3] *= (91);
      e[s][3] /= 100;
      e[s][n[s]-4] *= 91;
      e[s][n[s]-4] /= 100;
      e[s][4] *= (92);
      e[s][4] /= 100;
      e[s][n[s]-5] *= 93;
      e[s][n[s]-5] /= 100;
      }
  }
  }


  tt = 0;
  for (s=0; s<nstrings; s++) {
      tt += (int) ( amp *1* PREMUL * atan(
            ((double) e[s][30] + e[s][31] + e[s][32]) / (double) (amp * PREMUL * 10000)
         ));
  }
  tt *= 3;
  tt /= 9;

  for (s=0; s<nstrings; s++) {
   /* if (uc>s*400) { */
   if (midinote[s] > 0) {
      e [s][n[s]-3]  +=  4 * tt;
      e [s][n[s]-4]  +=  7 * tt;
      e [s][n[s]-5]  += 35 * tt;
      e [s][n[s]-6] += 160 * tt;
      e [s][n[s]-7] += 60 * tt;
      e [s][n[s]-8] +=  7 * tt;
      e [s][n[s]-9] +=  3 * tt;
   } else {
      e [s][n[s]-3]  +=  3 * tt;
      e [s][n[s]-4]  +=  5 * tt;
      e [s][n[s]-5]  += 8 * tt;
      e [s][n[s]-6] += 10 * tt;
      e [s][n[s]-7] += 7 * tt;
      e [s][n[s]-8] +=  6 * tt;
      e [s][n[s]-9] +=  2 * tt;
   }
  }
  /* uc++; */

  return;
}



simulate(n,tn,p)
int n[], tn;
short *p;
{
   int t;
   int r;
   for (t=0; t<tn; t++) {
      /* propagate(n); */
      propagate_fixend(n);
      if (r++ > 10) {          /* Try socket 1 in 10 times */
    /*     sockproc(10); */
         r=0;
      }
      outputpa(binoutnode,p);
      p+=2;
   }
}


closeup()
{
   if (fd > 0) close(fd);
}


process_args(argc,argv)
int argc;
char *argv[];
{
   int i;

   for (i=0; i< MAXNSTRINGS; i++)
     vel[i] = 500;

   printf("nocygwin version, doesn't need cygwin1.dll\n");
   printf("it does need glut32.dll\n");
   printf("click on one of the graphical window items for (reactivating) a string\n");
   printf("or push one of the numberical keys (1- #strings, 0 silences strings)\n");
   printf("\n");
   printf("author: M.T. Verelst, theover@yahoo.com \ndocu: http://195.241.128.75/articles/pms.html\n");

   if (argc <= 1) {
   
      nstrings = 3;
      n[0] = 100; n[1] = 150; n[2] = 75;
      binoutnode = 40;
      return; 
   }
   
   if ((argc<3) || (sscanf(argv[1],"%d",&nstrings) <= 0) ||
        nstrings < 1 || nstrings > MAXNSTRINGS ||
       (argc < nstrings+2) || ( ((argc-(nstrings+2))%2) != 0 )) {
      printf("Usage %s <#strings> <#nodes s1> .. <#nodes smax> [<option> <value>]...\n",
         argv[0]);
      printf("options: outnode <node #>   \n");
      exit(-1);
   }
   for (i=2; i<=nstrings+1; i++)
      sscanf(argv[i],"%d",&n[i-2]);
   /* sscanf(argv[2],"%d",&n[1]); */
/*   sscanf(argv[nstrings+2],"%d",&steps); */
   for (i=nstrings+2; i<argc; i++) {
      if (strcmp(argv[i],"soundfile") == 0) {

/*           initwav(n,argv[i+1]); i++; */
      } else 
      if (strcmp(argv[i],"tcl") == 0) {
         if (argv[i+1][0] == '1') notcl=0; else notcl=1; i++; 
      } else
      if (strcmp(argv[i],"initstate") == 0) {
         printf("Warning: initstate not yet implemented\n"); i++;
      } else
      if (strcmp(argv[i],"outnode") == 0) {
         sscanf(argv[i+1],"%d",&binoutnode); i++;
      } else {
         printf("Error: unknown option %s\n",argv[i]);
         exit(-1);
      }
   } 
}

int val;
char name[1024];

sockproc(r)
int r;
{
   int rr,i;
   static int count = 0;

   count+=r;

   while (  (rr = do_read()) > 0) {
      line[strlen(line)-1] = '\0';
      printf("set lastread {"); printf("%s",line); printf("}\n");
      fflush(stdout);
      if (strncmp(line,"val",strlen("val")) == 0) {
         sscanf(&line[strlen("val")],"%d",&val);
      }
      else
      if (strncmp(line,"tcl",strlen("tcl")) == 0) {
         char te[1024];
         sprintf(te,"set nnodes %d\n",n[0]);
         do_write(te);
         for (i=0; i<n[0]; i++) {
            sprintf(te,"set a(%d) %d\n",i,e[0][i]/PREMUL);
            do_write(te);
         }
         do_write("graphdel\n");
         sprintf(te,"graph $nnodes\nupdate\nafter $vt\n");
         do_write(te);
/*         sscanf(&line[strlen("tcl")],"%s",name);
         if (strlen(name) == 0) strcpy(name,"onewave.tcl");
         tclf = fopen(name,"w");
         foutputtcl(n);
         fflush(tclf);
         fclose(tclf); */
      }
      if (strncmp(line,"cur",strlen("cur")) == 0) {
         char t[1024];
         sprintf(t,"set iteration_count %d\n",count);
         fprintf(stdout,t);
         fflush(stdout);
         do_write(t);
      }
   }
}

writegraph(name)
char *name;
{
         tclf = fopen(name,"w");
         foutputtcl(n);
         fflush(tclf);
         fclose(tclf);
}

pms_do(k)
int k;
{
   if (k>0 && k<=nstrings)
      init(n,k-1);
   if (k==0) 
      init(n,-2);
}

void CALLBACK midiCallback(handle, uMsg, dwInstance, dwParam1, dwParam2)
HMIDIIN handle;
UINT uMsg;
DWORD dwInstance;
DWORD dwParam1;
DWORD dwParam2;
{
   int a,b,c, i;
   if (uMsg == MIM_OPEN) {
      printf("Midi interface callback function: device opened\n");
      return;
   }

   a = (dwParam1>> 0) & 0x0ff ;
   b = (dwParam1>> 8) & 0x0ff ;
   c = (dwParam1>>16) & 0x0ff ;
   if (a != 248 && a!=254)
/*      printf("midi mess. timer: %d event %d para1 %d para2 %d\n", dwParam2,
      a,b,c);  */

   if (a == 144 && c!=0) {
      i=0; while (i<MAXNSTRINGS && midinote[i] > 0) i++;
      if (i >= MAXNSTRINGS) i=0;
      /* n[i] = 3+101* pow(2,-(b-57.0)/12.0); */
      n[i] = 3+101* pow(2,-(b-57.0-12+2*bend)/12.0);
      vel[i] = c*c/18;
      pms_do(i+1);   /* trigger string  */
/*      printf("%d\n",i); */
      midinote[i] = b;
   }
   if (a == 128 || (a==144 && c==0)) {
      /* n[0] = b; */
      /* pms_do(0);   * silence all strings */
      for (i=0; i< MAXNSTRINGS; i++)
         if (midinote[i] == b) { 
	    /* silenceone(i); */
	    midinote[i] = -1;
         }
   }
   if (a == 224) {
      bend = (c-64.0)/64.0;
      for (i=0; i< MAXNSTRINGS; i++)
         if (midinote[i] > 0)
            n[i] = 3+101* pow(2,-(midinote[i] -57.0-12+2*bend)/12.0);
   }
   if (a == 176) {
      amp = c/64.0;
   }
   if (a == 176 && b == 7) {
      volume = c;
      volumechanged();
   }

   return;
}


init_midi()
{
  int i;
  printf("Midiopen: %d, ", midiInOpen(&inHandle, 0, (DWORD) midiCallback, 0, CALLBACK_FUNCTION));
  printf("midiInStart: %d.\n", midiInStart(inHandle));

  for(i=0; i< MAXNSTRINGS; i++)
     midinote[i] = -1;

   volume = 64;

}

main(argc, argv)
int argc;
char *argv[];
{
   notcl = 1;
   uc=100000;
   pick_main(&argc,argv);
   process_args(argc,argv);
   init(n, -1);
   serv_main(); 
   inittcl(n); 

   pa_maini(steps);   /* start portable sound library */
   init_midi();
   pick_loop();
   pa_maine();
   /* simulate(n,steps); */
   pick_end();
   closeup();
   return 0;
}

